/***************************************************************************
*
*   Module:            
*
*   Filename:           fruerror.c
*
*   Author:             Brian Tran
*
*   Created:            04/25/2005
*
*   Description:
*  
*   Routines to support run-time errors
*
****************************************************************************
*                 Source Control System Information
*
*    $Id: fruerror.c,v 1.2 2005/04/30 03:43:53 brian Exp $
*
****************************************************************************
*                       Modification History
*
*    $Log: fruerror.c,v $
*    Revision 1.2  2005/04/30 03:43:53  brian
*    Bugzilla number:
*    Add routine to add new entry without marking it new. Used for syncing between 2 cores
*
*    Revision 1.1  2005/04/27 09:32:26  brian
*    Bugzilla number:
*    Support runtime error reporting
*
*
***************************************************************************
*                       (c) COPYRIGHT 2000, 2001, 2002
*                           TopSpin Communications, Inc.
*                           Palo Alto, CA
**************************************************************************/
#ifdef RCSID
static const char rcsid[] = "$Id: fruerror.c,v 1.2 2005/04/30 03:43:53 brian Exp $";
#endif


#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/poll.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/init.h>
#include <linux/blk.h>
#include <linux/unistd.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/machdep.h>
#include <asm/ocp.h>
#include <linux/delay.h>
#include <i2c-errcode.h>
#include <topspin_types.h>
#include <fpga.h>
#include <system_interrupt_code.h>
#include "fruerror.h"


struct semaphore fruerr_lock[128];


/*
 * init_fruerror
 * -------------
 * Init data structure for a FRU. Call only once during startup
 */
void init_fru_error(int fru_location, tRUNTIME_FRU_ERROR *fruerror)
{
    fruerror->err_count = 0;
    fruerror->unprocessed_err_count = 0;
    fruerror->err_list = NULL;
    init_MUTEX(&fruerr_lock[fru_location]);

    return;
}


/*
 * reset_fruerror
 * -------------
 * Reset data structure for a FRU. Clear all current errors.
 */
void reset_fru_error(int fru_location, tRUNTIME_FRU_ERROR *fruerror)
{
    tRUNTIME_ERR_ENTRY *entry_pointer;
    tRUNTIME_ERR_ENTRY *entry_pointer_next;

    entry_pointer = fruerror->err_list;

    down(&fruerr_lock[fru_location]);

    fruerror->err_count = 0;
    fruerror->unprocessed_err_count = 0;
    while (entry_pointer != NULL) {
        entry_pointer_next = entry_pointer->next_err_entry;
        kfree(entry_pointer);
        entry_pointer = entry_pointer_next;
    }
    fruerror->err_list = NULL;

    up(&fruerr_lock[fru_location]);

    return;
}

/* 
 * make_err_entry
 * --------------
 * Allocate a space for a new entry
 * and copy data to it.
 */
tRUNTIME_ERR_ENTRY *make_err_entry(tRUNTIME_ERR_DATA entrydata)
{
    tRUNTIME_ERR_ENTRY *p;

    p = kmalloc (sizeof (tRUNTIME_ERR_ENTRY), GFP_KERNEL);
    if (p) {
        p->entrydata = entrydata;
        p->processed = 0;
        p->next_err_entry = NULL;
    }
    return p;
}

/*
 * found_match_in_fru_error
 * ------------------------
 * Given the errcode_key, check the list to see if we already have this error record.
 * Return 1 if match. Return 0 if not.
 * Also return a pointer to the matched record and a pointer to the record before the matched record.
 * If not match, either the list is empty or the list is not empty.
 * If the list is empty, both returned pointers are NULL.
 * If the list is not empty, matched_pointer is NULL and prev_matched_pointer is the end of the list.
 */
int found_match_in_fru_error(tRUNTIME_FRU_ERROR *fruerror, 
                             unsigned short errcode_key, 
                             tRUNTIME_ERR_ENTRY **matched_pointer,   
                             tRUNTIME_ERR_ENTRY **prev_matched_pointer)
{
    tRUNTIME_ERR_ENTRY *p;

    p = fruerror->err_list;
    *matched_pointer = NULL;
    *prev_matched_pointer = NULL;

    while (p) {
        if (p->entrydata.errcode_key == errcode_key) {
            *matched_pointer = p;
            return 1;
        }
        /* not match, move ahead */
        *prev_matched_pointer = p;
        p = p->next_err_entry; 
    }

    return 0;
}


   

/* 
 * add_entry_to_fru_error
 * ----------------------
 * Given the error code and error data, search the fru eror list
 * to see if this is an existing error.
 * If yes, update this entry with new error data and mark it not processed.
 * If no, create a new entry and apend it to the end of the list.
 */
void add_entry_to_fru_error(int fru_location, 
                            tRUNTIME_ERR_DATA entrydata,
                            tRUNTIME_FRU_ERROR *fruerror)
{

    tRUNTIME_ERR_ENTRY *matched_pointer;
    tRUNTIME_ERR_ENTRY *prev_matched_pointer;
    tRUNTIME_ERR_ENTRY *new_entry;

    down(&fruerr_lock[fru_location]);

    if (found_match_in_fru_error(fruerror, entrydata.errcode_key, &matched_pointer, &prev_matched_pointer)) {
        matched_pointer->processed = 0;
        matched_pointer->entrydata = entrydata;
        fruerror->unprocessed_err_count ++;
    }
    else {
        new_entry = make_err_entry(entrydata);
        if (prev_matched_pointer) {
            prev_matched_pointer->next_err_entry = new_entry;
        }
        else {
            fruerror->err_list = new_entry;
        }
        fruerror->err_count ++;
        fruerror->unprocessed_err_count ++;
    }

    up(&fruerr_lock[fru_location]);

    return;
}


/*
 * add_entry_to_fru_error_processed
 * ----------------------
 * Given the error code and error data, search the fru eror list
 * to see if this is an existing error.
 * If yes, update this entry with new error data.
 * If no, create a new entry and apend it to the end of the list.
 * Mark this node processed.
 */
void add_entry_to_fru_error_processed(int fru_location,
                            tRUNTIME_ERR_DATA entrydata,
                            tRUNTIME_FRU_ERROR *fruerror)
{

    tRUNTIME_ERR_ENTRY *matched_pointer;
    tRUNTIME_ERR_ENTRY *prev_matched_pointer;
    tRUNTIME_ERR_ENTRY *new_entry;

    down(&fruerr_lock[fru_location]);

    if (found_match_in_fru_error(fruerror, entrydata.errcode_key, &matched_pointer, &prev_matched_pointer)) {
        matched_pointer->entrydata = entrydata;
    }
    else {
        new_entry = make_err_entry(entrydata);
        new_entry->processed = 1;
        if (prev_matched_pointer) {
            prev_matched_pointer->next_err_entry = new_entry;
        }
        else {
            fruerror->err_list = new_entry;
        }
        fruerror->err_count ++;
    }

    up(&fruerr_lock[fru_location]);

    return;
}



/*
 * remove_entry_from_fru_error
 * ----------------------
 * Given the error code, search the fru error list
 * to see if this is an existing error.
 * If found AND this node is marked processed, remove this node.
 * Otherwise, ignore.
 * This routine is intended to remove a recovery event that has been processed.
 */
void remove_entry_from_fru_error(int fru_location,
                            unsigned short errcode_key,
                            tRUNTIME_FRU_ERROR *fruerror)
{

    tRUNTIME_ERR_ENTRY *matched_pointer;
    tRUNTIME_ERR_ENTRY *prev_matched_pointer;

    down(&fruerr_lock[fru_location]);

    if (found_match_in_fru_error(fruerror, errcode_key, &matched_pointer, &prev_matched_pointer)) {
        if (matched_pointer->processed) {
            if (prev_matched_pointer == NULL) {		/* we are at the head */
                fruerror->err_list = matched_pointer->next_err_entry;
            }
            else {
                prev_matched_pointer->next_err_entry = matched_pointer->next_err_entry;
            }
            kfree(matched_pointer);
            fruerror->err_count --;
        }
    }

    up(&fruerr_lock[fru_location]);

    return;
}


/*
 * process_entry_from_fru_error
 * ----------------------
 * Search the fru eror list for the next unprocessed entry.
 * Return 1 if there is an unprocessed entry, and return the data in this entry.
 * Return 0 otherwise.
 */
int process_entry_from_fru_error(int fru_location,
                            tRUNTIME_ERR_DATA *entrydata,
                            tRUNTIME_FRU_ERROR *fruerror)
{
    tRUNTIME_ERR_ENTRY *p;

    down(&fruerr_lock[fru_location]);

    if (fruerror->unprocessed_err_count == 0) {
        up(&fruerr_lock[fru_location]);
        return 0;
    }

    /* get the first unprocessed entry in the list */
    p = fruerror->err_list;    
    while (p) {
        if (p->processed == 0) {
            p->processed = 1;
            *entrydata = p->entrydata;
            up(&fruerr_lock[fru_location]);
            return 1;
        }
        p = p->next_err_entry;
    }

    up(&fruerr_lock[fru_location]);

    return 0;
}


/*
 * retrieve_all_processed_entry
 * ----------------------------
 * Go thru the list and return all entry data that have been processed.
 * This routine is intended for showing all fru errors.
 * Unprocessed entries will not be returned because thet need to be processed first
 * to be known to the system.
 * Entry data will be returned in the provided buffer. The size of the buffer is specified in max_num_entries.
 * The real numner of returned entries  will be returned in returned_num_entries. 
 */
void retrieve_all_processed_entry(int fru_location,
                                  int max_num_entries,
                                  int *returned_num_entries,
                                  tRUNTIME_ERR_DATA *entrydata,
                                  tRUNTIME_FRU_ERROR *fruerror)
{
    tRUNTIME_ERR_ENTRY *p;
    int count = 0;

    *returned_num_entries = 0;
    if (max_num_entries <= 0) {
        return;
    }

    down(&fruerr_lock[fru_location]);

    p = fruerror->err_list;   
    while (p) {
        if (p->processed == 0) {
            p = p->next_err_entry;
            continue;
        }
        *(entrydata + count) = p->entrydata;
        count ++;
        *returned_num_entries = count;
        if (count == max_num_entries) {		/* buffer is full */
            break;
        }
        p = p->next_err_entry;
    }

    up(&fruerr_lock[fru_location]);

    return;
}

/*
 * make_fru_error
 * ---------------
 * Make a new fru error list based on the data ptovided.
 * Mark all entries processed.
 * This routine is intended to be used for data syncing bwtween 2 cores.
 */
void make_fru_error(int fru_location,
                    int num_entries,
                    tRUNTIME_ERR_DATA *entrydata,
                    tRUNTIME_FRU_ERROR *fruerror)
{      
    tRUNTIME_ERR_ENTRY *new_entry;
    tRUNTIME_ERR_ENTRY *tail;
    int i;

    if (num_entries <= 0) {
        return;
    }

    /* first reset the list */
    reset_fru_error(fru_location, fruerror);

    down(&fruerr_lock[fru_location]);
    new_entry = make_err_entry(*entrydata);
    /* mark this entry processed */
    new_entry->processed = 1;
    tail = new_entry; 
    fruerror->err_list = new_entry; 
    fruerror->err_count = 1; 
    fruerror->unprocessed_err_count = 0; 
    for (i = 1; i < num_entries; i++ ) {
        new_entry = make_err_entry(*(entrydata + i));
        new_entry->processed = 1;
        fruerror->err_count ++;
        tail->next_err_entry = new_entry;
        tail = new_entry;
    }
    up(&fruerr_lock[fru_location]);

    return;

}













