/*
  Boot from M-System DiskOnChip Millenium Plus

  Copyright (c) 2002 Topspin Communications
*/

#include <ppcboot.h>
#include <malloc.h>
#include <command.h>
#include <cmd_mboot.h>
#include <net.h>

#include <blockdev.h>

extern unsigned int sread(volatile unsigned int *addr);

#if (CONFIG_COMMANDS & CFG_CMD_MBOOT)

/* 
 * Drive Handle
 *
 * 7 - 4 			3 - 0
 * Disk Partition	Physical Drive
 * 
 * An M-Systems Disk Partition refers to the BDTL partition
 * used by dformat to format the drive.
 * The physical drive is the physical drive number of the DiskOnChip
 * device (usually 0).
 * 
 * The Topspin "main disk partition" where we usually boot from is:
 *
 * Disk Partition = 0, Physical Drive = 0  ie. = 0x00
 *
 * The Topspin "recovery disk partition" is:
 *
 * Disk Partition = 1, Physical Drive = 0   ie. = 0x10
 *
 */

static int boot_handle;

/* boot block, partition info */
static Bootblock_x86_t bootblock_info ;
static Partition_info_t part_info ;

/* TFFS driver init flag */
static u_char   initialized = 0 ;

/* Early malloc memory area */
unsigned char *early_malloc_p = (unsigned char *)CFG_INIT_MALLOC_ADDR;

#define BLOCK_SIZE          512
static int get_blk(int blk, u_char *buff) ;

int mdocp_init(void)
{
  FLStatus  status ;
  IOreq     ioreq;
  char *s;
  volatile unsigned int bootsig;

  s = getenv ("bootpart");
  /* 
   * If the boot signature location contains the "recovery image" boot 
   * signature, boot the recovery image from disk part 1. Otherwise
   * boot the normal image from disk part 0.
   */
  if (s == NULL) {
      bootsig = sread(TS_BOOTSIG_ADDR);
      if (bootsig == TS_BOOTSIG_RECOVERY) {
          boot_handle = 0x10;
          printf ("\nBoot Recovery Image from Disk Partition 1: boot sig=0x%x\n",
              bootsig); 
      } else {
          boot_handle = 0x00;
          printf ("\nBoot Regular Image from Disk Partition 0: boot sig=0x%x\n",
              bootsig); 
      }

  } else if (strcmp(s, "1") == 0) {   /* boot disk part 1 (recovery image) */
      boot_handle =  0x10;
      printf ("\nEnv bootpart=1: Boot Recovery Image from Disk Partition 1\n"); 
  } else if (strcmp(s, "0") == 0) {    /* boot disk part 0 (normal image) */
      boot_handle =  0x00;
      printf ("\nEnv bootpart=0: Boot Regular Image from Disk Partition 0\n");
  }


#ifdef DEBUG
  printf ("Boot Drive: M-Systems Disk Partition:BDTL%d Phy Drive:%d\n", (boot_handle>>4), (boot_handle&&0x0f));
#endif

  /* initializing only ONCE */
    if (initialized == FALSE)
    {
        early_malloc_p = NULL;

#ifdef DEBUG
        printf ("\nInitializing TFFS:") ;
        printf ("\n  - flInit() -> ") ;
#endif
        status = flInit() ;
        if (status != flOK)
          return -1 ;
#ifdef DEBUG
        printf ("ok") ;
#endif

        memset((char*)&ioreq, 0, sizeof(IOreq)) ;
#ifdef DEBUG
        printf ("\n  - Mount the TFFS volume -> ") ;
#endif

        /* Init and Mount the volume */
        ioreq.irHandle = boot_handle;
        status = flAbsMountVolume(&ioreq);
        if ((status == flOK) || (status == flNonFATformat))
        {
#ifdef DEBUG
            printf ("ok") ;
            printf ("\n  - Reading master boot sector -> ") ;
#endif
            /* read the bootblock info */
            if (get_blk(0, (char*)&bootblock_info) != 1)
                return -1 ; /* can't read boot block */

#ifdef DEBUG
            printf ("ok") ;
            printf ("\n  - Verify the master boot sector magic number  -> ") ;
#endif
            if ( (bootblock_info.boot_magic[0] != 0x55) ||
               (bootblock_info.boot_magic[1] != 0xaa) )
            return -1 ; /* no magic */
#ifdef DEBUG
            printf ("ok") ;
            printf ("\n") ;
#endif
            initialized = TRUE ;
	    }
        else
          return -1;
    }

    /* set partition info */
    part_info.start = LE4((char*)&bootblock_info.partitions[0].first_block);
    part_info.size = part_info.raw_size = LE4((char*)&bootblock_info.partitions[0].size);
    part_info.in_use = 1 ;
    
    /* Print out the partiotions information */
#ifdef DEBUG
    printf("  - Partition start=%d,size=%d\n", (int)part_info.start, (int)part_info.size);
#endif

    return 0;
}    


int do_mboot (cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    uint32_t bno;
    uint32_t offset = 0;

#ifdef DEBUG
    printf("Booting from Disk OnChip Millenium Plus\n");
#endif
    
    /* Process the command line arguments (if any) */
    if (argc > 2) {
            printf("Usage:\n%s\n", cmdtp->usage);
            return 1;
    }

    if (argc == 2)
        load_addr = simple_strtoul(argv[1], NULL, 16);
	
#ifdef DEBUG
    printf("Load address: 0x%08lx\n", load_addr);
#endif
    
    eth_halt();
    eth_init(bd);
    eth_halt();

    /* Init the TFFS */    
    if (mdocp_init())
    {
        printf("Failed to initialize the M-Systems TFFS. Aborted !\n");
        return 1;
    }
    
    printf("Loading:      ");

    /* Copy the kernel image partition into the load area */
    for (bno = part_info.start; bno < part_info.size ; bno++, offset += BLOCK_SIZE )
    {
        if (get_blk(bno, (char*)(load_addr + offset)) != 1)
        {
            printf("mboot: error reading block %d. Abort!\n", (int)bno);
            return -1 ;
        }
        if (bno % 64 == 0) {
            putc('+');
            if (bno % (64 * 40) == 0) {
                    printf("[%5dK]\n              ", offset / 1024);
            }
        }
    }

    while (bno % (64 * 40)) {
      ++bno;
      if (bno % 64 == 0) {
        putc(' ');
      }
    }
    printf("[ Done ]\n");
    
    return 0;
}

/* The following are the PPCBOOT versions of various things required by 
   the M-Systems OSAK that must be compiled here in the Linux kernel environment.
   These are prototyped in flsysfun.h which is NOT included here as it drags in 
   too much other proprietary data. */
void flDelayMsecs(unsigned milliseconds)
{
        register int    i;
        for (i = 0; i < milliseconds; i++)
                udelay(1000L);
}

void flsleep(unsigned long milliseconds)
{
    flDelayMsecs(milliseconds);
}

void*flmemcpy(void *dest, const void *src, size_t length)
{
        return memcpy(dest, src, length);
}

void*flmemset(void *dest, int value, unsigned int length)
{
        return memset(dest, value, length);
}

void *flmemcpy_fromio(void *dest, const void * src, unsigned int length)
{
        return(memcpy(dest, src, length));
}

void *flmemcpy_toio(void *dest, const void * src, unsigned int length)
{
        return(memcpy(dest, src, length));
}

void*flmemset_io(void *dest, int value, unsigned int length)
{
        return memset(dest, value, length);
}

int flmemcmp(const void * dest,const void  *src,size_t count)
{
        int res;
        res = memcmp(dest, src, count);
        return(res);
}

static unsigned long fl_winl=0xfff00000;
static unsigned long fl_winh=0xfff00000;

unsigned flRandByte(void)
{
        return (get_tbclk() >> 1) & 0xff;
}

unsigned long flGetWinL(void)
{
        return fl_winl;
}

unsigned long flGetWinH(void)
{
        return fl_winh;
}

void *early_malloc(unsigned long size)
{
  unsigned char *retp;

  if (early_malloc_p == NULL)
    return NULL;

  if ((unsigned long)early_malloc_p + size > CFG_INIT_MALLOC_ADDR + CFG_MALLOC_LEN)
    return NULL;

  retp = early_malloc_p;
  early_malloc_p += size;

  return  retp;
} /* early_malloc */

void*flvmalloc(unsigned long size)
{
  if (early_malloc_p == NULL)
    return malloc(size);
  else
    return early_malloc(size);
}

void flvfree(void * addr)
{
  if (early_malloc_p == NULL)
    free(addr);
}

static int get_blk(int blk, u_char *buff)
{
  IOreq     ioreq;
  FLStatus  status ;

  if (buff == NULL)
    return FALSE ;

  memset((char*)&ioreq, 0, sizeof(IOreq)) ;
  ioreq.irHandle = boot_handle;
  ioreq.irSectorNo    = blk;
  ioreq.irSectorCount = 1;
  ioreq.irData        = buff;

  /* Read the requested sector */
  status = flAbsRead(&ioreq) ;
  if (status != flOK)
      return(FALSE);

  return (TRUE) ;
}

#endif /* CFG_CMD_MBOOT */

